/*
 * Store state updates in time as one history report. The history report is
 * one stack of entries. Each entry has one history fact, and is composed by
 * two elements: one label and one data (DATAPT). The label is a non-empty
 * string identifier (not necessarily unique) for the entry. Entry data is
 * used to store the state update.
 *
 * Entries are added to on top of stack. All previous entries are in
 * these stack, too.
 *
 * State updates are defined as requests to insert one entry in stack.
 *
 * This class was derived from class_list.
 * To know the number of entries use: class_list::get_total_nodes ().
 *
 * User can predefine some entries data using the discrete function
 * resource.
 * 
 * LICENSE:
 *
 * Copyright (C) 2000 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <stdlib.h>
#include "../../include/base/type.h"
#include "../../include/base/list.h"
#include "../../include/base/mapstr.h"
#include "../../include/base/history.h"



struct entry {
        char* label;
	DATAPT data;
};



class_history::class_history ()
{
	init ();
}



class_history::~class_history ()
{
  terminate ();
}



void  class_history::
init ()
{
  class_list::init ();
  predefined_data.init ();
}



/* Free all data, don't desallocate pointed areas
   from entries (if pointed) and predefined
   data. */
void  class_history::
terminate ()
{
  // empty stack
  entry *en;
  while (remove_first (put_in en)) {
    free (en);
  }
  class_list::terminate ();

  // empty predefined data map
  predefined_data.terminate ();
}



/* Update history adding one entry on top of stack.
The top is the first node (indexed as zero).

label - the entry label, can identify entry data type, or
        subject, priority, or something like that.
	The given string is refered (is not duplicated!) - don't
	deallocate it until this class instance is up.
	Label can't be an empty string.
data - entry data
select - if TRUE select this entry / FALSE - don't

return:
TRUE - updated / FALSE - don't
*/
BOOL class_history::update ( char* label, DATAPT data, BOOL select )
{
  // if label is empty, abort
  if ( label == 00 || label [0] == 0 )
    return FALSE;

  entry* en = (entry*) malloc ( sizeof (entry) );
  if (!en) return FALSE;
  
  en->label = label;
  en->data = data;

  // add entry in stack
  if ( ! insert_first ( BEFORE, get_from en ) ) {
    free ( en );
    return FALSE;
  }

  // select, if this is the case
  if ( select )
    select_first ();

  return TRUE;
}



/* Predefine one data entry. User must give one
   key (string) to be used as data identifier.

   key - key for the predefined entry data
   data - the entry data

   return:
   TRUE - ok / FALSE - error
 */
BOOL class_history::predefine_data ( char* key, DATAPT data )
{
  // if key is empty, abort
  if ( key == 00 || key [0] == 0 )
    return FALSE;

  return predefined_data.insert ( key, data );
}



/*  Remove the first predefined data from the table.
 The order is pointless because this function is just designed
 to empty the map interactively with some user loop.

 return:
 data - the data
 point the associated key, or NULL if no elements found.
 */
char* class_history::
remove_any_predefined_data ( DATAPT& data)
{
  char* key = predefined_data.get_at_position ( 0, data );
  if ( !key ) return 00;

  return predefined_data.remove ( key, data );
}



/*  Remove the predefined data with the given key from the table.
    key - key of element to be removed
 return:
 data - the data
 point the associated key, or NULL if no elements found.
 */
char* class_history::
remove_predefined_data ( char* key, DATAPT& data)
{
  return predefined_data.remove ( key, data );
}



/* Remove the selected entry.
   return:
   TRUE - ok / FALSE - error,
   probably there hasn't selection
*/
BOOL class_history::
remove_selected_entry ()
{
  DATAPT nothing;
  return remove_selected ( put_in nothing );
}



/* Get the predefined data associated with the given key.
   key - string used as data identifier
   return:
   data - associated data
   TRUE - ok / FALSE - data not found
 */
BOOL class_history::
get_predefined_data ( char* key, DATAPT& data )
{
  DWORD num = predefined_data.get_value ( key, data );
  if ( num )
    return TRUE;
  else
    return FALSE;
}



/* Get the selected entry data and label.
   return:
   data - the entry data
   pointer to entry label or 00 if there was no selection
*/
char* class_history::
get_selected_entry ( DATAPT& data )
{
  entry *en;
  BOOL res = get_selected ( put_in en );
  if ( !res ) return 00;

  data = en->data;
  return en->label;
}



/* Get data from the nth match entry with the given label,
   or only the data from the entry in the given position.

   pos - nth position (start at index 0)
   label - label to search for
   select - if TRUE select the entry found or makes selection
            none if entry wasn't found. If FALSE don't touch
            in selection.

   If label is 00 then pos is used as indexing all
   history entries. If label is not 00, then pos refer
   to the nth entry with the same label.

   return:
   data - the given data
   the entry label associated with data, or 00 if entry not found
*/
char* class_history::
get_entry ( DWORD pos, char* label, BOOL select, DATAPT& data)
{
  // for each entry in history
  BOOL end = FALSE;
  DWORD hist_pos = 0;       // entry position in history
  entry *en;
  while ( ! end ) {

    BOOL res;
    // if is to select entry
    if ( select ) {
      select_at_position ( hist_pos );
      res = get_selected ( put_in en );
    }      
    else
      res = get_at_position ( hist_pos, put_in en );
    
    // end of history - entry not found
    if ( !res ) {
      if ( select ) select_none ();
      return 00;
    }

    // compare labels, if was defined one label to look for
    BOOL match = FALSE;
    if ( label ) {
      if ( strcmp ( en->label, label ) == 0 )
	// labels match!
	match = TRUE;
    }
    else
      // was not defined the given label
      match = TRUE;
    
    // if found one entry with the given label
    if ( match )
      // verify if it's in the right position
      if ( pos == 0 ) {
	// yes, it's
	data = en->data;
	return en->label;
      }
      else // no, it is not
	pos --;
    
    hist_pos ++;
  }

  if ( select ) select_none ();
  return 00;
}
