/*
 * Implement a double linked list.
 * List is composed by elements called nodes.
 * 
 * FEATURES:
 * 
 * possible operations on list: insert or remove nodes, access nodes
 * data, select one specific node. Many of these operations has specific
 * functions to manipulate nodes in: first, last, midle or selected positions.
 * Other operations can be apllyed over the hole list: sort, make list
 * circular or linear, return the number of nodes, append, split and revert
 * the list, search for nodes.
 * 
 * direct access: this feature make possible fast access for nodes in a
 * random position. Is implemented with an array of pointers to list nodes.
 * See CAUTION.
 *
 * List can be sorted acording node data, see sort () for details.
 * See find_sorted () and find_unsorted () for explanations about how to search
 * nodes in list.
 * 
 * TIPS:
 *
 * to know if selection operations were complited successfully make this test:
 * if ( has_selection () ) {...ok...}
 * 
 * class_node is the nodes class.
 * Use its data member variable called 'data' to store information.
 * 
 * to know if list is empty use the comparison:
 * get_total_nodes () == 0
 * 
 * the index of first node is: 0
 * 
 * the index of last node is: get_total_nodes () - 1
 *
 * when zero is assigned to a pointer is used the sysntax 00 
 * in place of the most common 0. This is just a convension, not really necessary.
 * 
 * CAUTION:
 *
 * direct access is not updated automaticaly if you change the list
 * pointers (with insertion or remotion of nodes). Before change nodes call
 * class_list::disable_direct_access (), and after (if you still want direct access)
 * call class_list::enable_direct_access ().
 * 
 * THE DESTRUCTOR DOES NOT DELETE (desalocate) NODES.
 * YOU MUST CALL remove_all () BEFORE DROP LIST OBJECT.
 * 
 * LICENSE:
 *
 * Copyright (C) 1999 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <stdlib.h>
#include "../../include/base/type.h"
#include "../../include/base/list.h"



DATAPT default_compare_function (DATAPT data1, DATAPT data2);



/*
 * constructor - init class variables
 */
class_list::class_list ()
{
	init ();
}



/*
 * destructor
 */
class_list::~class_list ()
{
	terminate ();
}



/*
 * set default values
 */
void  class_list::
init ()
{
	m_first = 00;
	m_last = 00;
	m_selected = 00;
	m_total_nodes = 0;
	m_direct_access = 00;
	compare_function = default_compare_function;
}



/*
 * Actions done when terminateing the class instance.
 */
void  class_list::
terminate ()
{
}



/*
 * Remove all nodes.
 * delete_data - if TRUE: deallocate node data memory, if FALSE - no
 * Default for delete_data is FALSE
 */
void  class_list::
remove_all (BOOL delete_data)
{
	// remove all nodes
	for (DWORD count = 0; count < m_total_nodes; count++) {

		class_node *node = m_first;

		// find next node
		m_first = m_first->next;

		// if need, deallocate data
		if (delete_data)
			free ((void *) get_data (node));

		// deallocate node
		free_node (node);
	}

	disable_direct_access ();

	m_first = 00;
	m_last = 00;
	m_selected = 00;
	m_total_nodes = 0;
	m_direct_access = 00;
}



/*
 * Remove the first node.
 *
 * return:
 * data - the node data
 * TRUE - if node deleted / FALSE - if not (node not found)
 */
BOOL  class_list::
remove_first (DATAPT & data)
{
	// if list is empty
	if (!m_first)
		return FALSE;

	class_node *node = m_first;
	remove_node (node);
	data = get_data (node);
	free_node (node);

	return TRUE;
}



/*
 * Remove the last node.
 *
 * return:
 * data - the node data
 * TRUE - if node deleted / FALSE - if not (node not found)
 */
BOOL  class_list::
remove_last (DATAPT & data)
{
	// if list is empty
	if (!m_last)
		return FALSE;

	class_node *node = m_last;
	remove_node (node);
	data = get_data (node);
	free_node (node);

	return TRUE;
}



/*
 * Remove the selected note, select the previous
 * (or next if previous don't exist).
 *
 * return:
 * data - the node data
 * TRUE - if node deleted / FALSE - if not (node not found)
 */
BOOL  class_list::
remove_selected (DATAPT & data)
{
	// if don't exist a selected node
	if (!m_selected)
		return FALSE;

	class_node *node = m_selected;
	remove_node (m_selected);
	data = get_data (node);
	free_node (node);

	return TRUE;
}



/*
 * Remove node in the given position (index).
 *
 * return:
 * data - the node data
 * TRUE - if node deleted / FALSE - if not (node not found)
 */
BOOL  class_list::
remove_at_position (DWORD index, DATAPT & data)
{
	// find current node in the given position (index)
	class_node *node = get_node_at_position (index);
	if (!node)
		return FALSE;

	remove_node (node);
	data = get_data (node);
	free_node (node);

	return TRUE;
}



/*
 * Insert one node after (or before) first node.
 *
 * local - first node relative location to make insertion,
 *         possible values are: AFTER or BEFORE
 * data - data for the new node
 *
 * return:
 * TRUE - success / FALSE - error (not enough memory)
 */
BOOL  class_list::
insert_first (relative_location local, DATAPT data)
{
	// try allocate node
	class_node *node = alloc_node;
	if (!node)
		return FALSE;

	insert_node (node, local, m_first);
	set_data (node, data);

	return TRUE;
}



/*
 * Insert one node after (or before) last node.
 *
 * local - last node relative location to make insertion,
 *         possible values are: AFTER or BEFORE
 * data - data for the new node
 *
 * return:
 * TRUE - success / FALSE - error (not enough memory)
 */
BOOL  class_list::
insert_last (relative_location local, DATAPT data)
{
	// try allocate node
	class_node *node = alloc_node;
	if (!node)
		return FALSE;

	insert_node (node, local, m_last);
	set_data (node, data);

	return TRUE;
}



/*
 * Insert node before or after the selected node.
 *
 * local - selected node relative location to make insertion,
 *         possible values are: AFTER or BEFORE
 * data - data for the new node
 *
 * return:
 * TRUE - success / FALSE - error (not enough memory)
 */
BOOL  class_list::
insert_selected (relative_location local, DATAPT data)
{
	// try allocate node
	class_node *node = alloc_node;
	if (!node)
		return FALSE;

	if (!m_selected) {
		insert_node (node, local, m_last);
		m_selected = 00;
	}
	else
		insert_node (node, local, m_selected);

	set_data (node, data);

	return TRUE;
}



/*
 * Insert one node after or before the node specified by index.
 *
 * If index out of range, insert at list end.
 *
 * local - location to make insertion, relative to node at the given index,
 *         possible values are: AFTER or BEFORE
 * data - data for the new node
 * index - position to insert
 *
 * return:
 * TRUE - success / FALSE - error (not enough memory)
 */
BOOL  class_list::
insert_at_position (relative_location local, DWORD index, DATAPT data)
{
	// find current node in the given position (index)
	class_node *current_node = get_node_at_position (index);
	if (!current_node)
		current_node = m_last;

	// try allocate new node
	class_node *node = alloc_node;
	if (!node)
		return FALSE;

	insert_node (node, local, current_node);
	set_data (node, data);

	return TRUE;
}



/*
 * Return: first node data and TRUE if node found, else FALSE.
 */
BOOL  class_list::
get_first (DATAPT & data)
{
	if (!m_first)
		return FALSE;

	data = get_data (m_first);

	return TRUE;
}



/*
 * Return: last node data and TRUE if node found, else FALSE.
 */
BOOL  class_list::
get_last (DATAPT & data)
{
	if (!m_last)
		return FALSE;

	data = get_data (m_last);

	return TRUE;
}



/*
 * Return: selected node data and TRUE if node found, else FALSE.
 */
BOOL  class_list::
get_selected (DATAPT & data)
{
	if (!m_selected)
		return FALSE;

	data = get_data (m_selected);

	return TRUE;
}



/*
 * Return: data in node specified by index and
 * TRUE if node found, else FALSE.
 */
BOOL  class_list::
get_at_position (DWORD index, DATAPT & data)
{
	class_node *node = get_node_at_position (index);
	if (!node)
		return FALSE;

	data = get_data (node);
	return TRUE;
}



/*
 * Select the first node.
 */
void  class_list::
select_first ()
{
	m_selected = m_first;
}



/*
 * Select the last node.
 */
void  class_list::
select_last ()
{
	m_selected = m_last;
}



/*
 * Select node in the given position (index).
 */
void  class_list::
select_at_position (DWORD index)
{
	m_selected = get_node_at_position (index);
	return;
}



/*
 * Define selected node as one of the next nodes acording the given offset.
 * Default is offset 1.
 */
void  class_list::
select_next (DWORD offset)
{
	// seek nodes
	for (DWORD pos = 0; pos < offset; pos++)
		// verify if was found the end of list
		if (m_selected->next == 00) {
			m_selected = 00;
			return;
		}
		else
			m_selected = m_selected->next;

	return;
}



/*
 * Define selected node as one of the previous nodes, acording the offset.
 * Default offset 1.
 */
void  class_list::
select_previous (DWORD offset)
{
	// seek nodes
	for (DWORD pos = 0; pos < offset; pos++)
		// verify if was found the begin of list
		if (m_selected->previous == 00) {
			m_selected = 00;
			return;
		}
		else
			m_selected = m_selected->previous;

	return;
}



/*
 * Change selection acording the position of selected node and
 * the given offset.
 */
void  class_list::
select_offset (int offset)
{
	// seek nodes
	if (offset < 0) {

		offset = -offset;

		for (int pos = 0; pos < offset; pos++)
			// verify if was found the end of list
			if (m_selected->previous == 00) {
				m_selected = 00;
				return;
			}
			else
				m_selected = m_selected->previous;
	}
	else
		for (int pos = 0; pos < offset; pos++)
			// verify if was found the begin of list
			if (m_selected->next == 00) {
				m_selected = 00;
				return;
			}
			else
				m_selected = m_selected->next;

	return;
}



/*
 * Makes selection null.
 */
void  class_list::
select_none ()
{
	m_selected = 00;
}



/*
 * Swap first node data with the given data.
 * 
 * return:
 * data - the old (replaced) data
 * TRUE if data replaced / FALSE if not (node not found)
 */
BOOL  class_list::
swap_first (DATAPT & data)
{
	if (!m_first)
		return FALSE;

	// swap
	DATAPT tmp = get_data (m_first);
	set_data (m_first, data);
	data = tmp;

	return TRUE;
}



/*
 * Swap last node data with the given data.
 * 
 * return:
 * data - the old (replaced) data
 * TRUE if data replaced / FALSE if not (node not found)
 */
BOOL  class_list::
swap_last (DATAPT & data)
{
	if (!m_last)
		return FALSE;

	// swap
	DATAPT tmp = get_data (m_last);
	set_data (m_last, data);
	data = tmp;

	return TRUE;
}



/*
 * Swap selected node data with the given data.
 * 
 * return:
 * data - the old (replaced) data
 * TRUE if data replaced / FALSE if not (node not found)
 */
BOOL  class_list::
swap_selected (DATAPT & data)
{
	if (!m_selected)
		return FALSE;

	// swap
	DATAPT tmp = get_data (m_selected);
	set_data (m_selected, data);
	data = tmp;

	return TRUE;
}



/*
 * Swap data for node in the given index with the given data.
 * 
 * return:
 * data - the old (replaced) data
 * TRUE if data replaced / FALSE if not (node not found)
 */
BOOL  class_list::
swap_at_position (DWORD index, DATAPT & data)
{
	class_node *node = get_node_at_position (index);
	if (!node)
		return FALSE;

	// swap
	DATAPT tmp = get_data (node);
	set_data (node, data);
	data = tmp;

	return TRUE;
}



/*
 * Append the given source list after or before first node.
 *
 ^ local - node relative location to make insertion,
 *         possible values are: AFTER or BEFORE
 */
void  class_list::
append_first (relative_location local, class_list & source)
{
	append_list (source, local, m_first);
}



/*
 * Append the given source list after or before last node.
 *
 ^ local - node relative location to make insertion,
 *         possible values are: AFTER or BEFORE
 */
void  class_list::
append_last (relative_location local, class_list & source)
{
	append_list (source, local, m_last);
}



/*
 * Append the given source list after or before selected node.
 *
 ^ local - selected node relative location to make insertion,
 *         possible values are: AFTER or BEFORE
 */
BOOL  class_list::
append_selected (relative_location local, class_list & source)
{
	// if list has nodes but none are selected, abort
	if (m_first && !m_selected)
		return FALSE;

	append_list (source, local, m_selected);

	return TRUE;
}



/*
 * Append the given source list after or before node at the given
 * position specified by index.
 *
 * If index out ou range, append at end.
 *
 ^ local - selected node relative location to make insertion,
 *         possible values are: AFTER or BEFORE
 */
void  class_list::
append_at_position (DWORD index, relative_location local, class_list & source)
{
	class_node *node = get_node_at_position (index);
	if (!node)
		node = m_last;

	append_list (source, local, node);
}



/*
 * Split list from node with index 'start_index' to node
 * with index 'end_index'.
 *
 * Warning:
 * The 'destin' list is cleaned before operation.
 * Don't fill it with nodes before split, they would the lost!
 *
 * This function is low optimized, use it only when necessary.
 *
 * return:
 * TRUE - success / FALSE - error (indexes out of range)
 */
BOOL  class_list::
split (DWORD start_index, DWORD end_index, class_list & destin)
{
	// verify if indexes are out of range
	if (start_index >= m_total_nodes || end_index >= m_total_nodes)
		return FALSE;

	// prepare destin to receive data
	destin.remove_all ();

	// sort indexes, start index must be less than (or equal with) end index
	if (start_index > end_index) {
		DWORD tmp = start_index;
		start_index = end_index;
		end_index = tmp;
	}

	// remove nodes from this list and put in destin list
	DWORD total_move = end_index - start_index + 1;
	for (DWORD ind = 0; ind < total_move; ind++) {

		class_node *node = get_node_at_position (start_index);
		remove_node (node);
		destin.insert_node (node, AFTER, destin.m_last);
	}

	return TRUE;
}



/*
 * Return TRUE if one node is selected, FALSE if selection is empty (00).
 */
BOOL  class_list::
has_selection ()
{
	if (m_selected == 00)
		return FALSE;

	return TRUE;
}



/*
 * Return number of nodes.
 */
DWORD class_list::
get_total_nodes ()
{
	return m_total_nodes;
}



/*
 * Define the index of selected node.
 *
 * This function is medium optimized, use only when necessary.
 *
 * return:
 * index - the selection index (if exist selection)
 * TRUE - selection exist / FALSE - selection is null
 */
BOOL  class_list::
get_select_position (DWORD & index)
{
	if (!m_selected)
		return FALSE;

	class_node *node = m_first;
	for (DWORD pos = 0; pos < m_total_nodes; pos++) {

		if (node == m_selected) {
			index = pos;
			return TRUE;
		}

		node = node->next;
	}

	return FALSE;
}



/*
 * Link the extreme pointers, doing a circular list.
 */
void  class_list::
do_circular ()
{
	// verify if list is not empty
	if (m_first != 00) {
		m_first->previous = m_last;
		m_last->next = m_first;
	}
}



/*
 * Unlink extreme pointers, doing list linear.
 */
void  class_list::
do_linear ()
{
	// verify if list is not empty
	if (m_first != 00)
		m_first->previous = m_last->next = 00;
}



/*
 * Create direct access array.
 * return TRUE - ok or list empty / FALSE, can't create
 */
BOOL  class_list::
enable_direct_access ()
{
	disable_direct_access ();

	// verify if index is out of range, if yes return
	if (m_total_nodes == 0)
		return TRUE;

	// try alocate memory for direct access array
	m_direct_access = (class_node **) malloc (sizeof (class_node *) * m_total_nodes);
	if (m_direct_access == 00)
		return FALSE;

	// navigate list, set array pointers to list nodes
	class_node *node = m_first;
	for (DWORD index = 0; index < m_total_nodes; index++) {

		m_direct_access [index] = node;

		// next node
		node = node->next;
	}

	return TRUE;
}



/*
 * Deallocate direct access array.
 */
void  class_list::
disable_direct_access ()
{
	if (m_direct_access != 00) {
		free (m_direct_access);
		m_direct_access = 00;
	}
}



/*
 * Sort list.
 *
 * Is necessary set first the comparison function
 * (see class_list::set_compare_function) to make sort work with user defined data.
 *
 * There is no need to update direct access after sort.
 *
 * The list is not automatically resorted when new nodes are inserted.
 *
 * Selected node data can change after sort.
 *
 * If list is sorted and a new node is inserted at end, its final
 * position after a resort is the end of nodes with equal results
 * in comparison. If a new one is inserted at list head, its final
 * position would be the begin of nodes with equal results.
 *
 * The sort can be ascending or descending acording the user comparison function.
 * Compared nodes has their data swapped if result is greater than 0.
 * List is sorted with bubble sort method.
 */
void  class_list::
sort ()
{
	// if list not empty
	if (m_total_nodes == 0)
		return;

	// sort list
	BOOL  sorted;
	do {

		// scan list
		sorted = TRUE;	// supose list is sorted

		class_node *node = m_first;
		//// this node would be compared with its next

		for (DWORD pos = 0; pos < m_total_nodes - 1; pos++) {

			// get data
			DATAPT data1 = get_data (node);
			DATAPT data2 = get_data (node->next);

			// compare data
			DATAPT comp_result = (*compare_function) (data1, data2);

			// analise result
			if (comp_result > 0) {	// data1 is greater than data2

				// swap

				DATAPT tmp = data1;
				data1 = data2;
				data2 = tmp;

				set_data (node, data1);
				set_data (node->next, data2);

				// list was not sorted
				sorted = FALSE;
			}

			node = node->next;
		}
	} while (!sorted);
}



/*
 * Find nodes with the given data using "bynary-tree search".
 *
 * Comparison between nodes and the given data is done with the user defined
 * compare function (see class_list::set_compare_function for details).
 *
 * If just one node was found, it's selected.
 *
 * In some cases more than one node can match. In those cases, parameter 'local'
 * define if the first or last match nodes would be selected, respectively, the
 * possible values are: BEFORE or AFTER.
 * 
 * If not found any equal node, is selected the closest node
 * after or before (acording 'local' parameter).
 *
 * Caution: (important note about struct pointers)
 * if 'data' parameter is a pointer to any other data structure
 * with more pointers, is important fill the fields used in nodes comparison
 * with usefull data and set to null (00) the other fields.
 *
 * List must be sorted before call.
 *
 * return:
 * number of nodes with match data
 */
DWORD class_list::
find_sorted (relative_location local, DATAPT data)
{
	DWORD num;
	m_selected = make_btree_search (local, data, num);
	return num;
}



/*
 * Find nodes with the given data using "linear search". List
 * need not be sorted, but this function is slowest than find_sorted.
 *
 * Comparison between nodes and the given data is done with the user defined
 * compare function (see class_list::set_compare_function for details).
 *
 * If some node was selected, search from its next/previous node,
 * acording 'local' value: AFTER or BEFORE, respectively to list extreme.
 *
 * Stop when the first match node was found, selecting it.
 *
 * If reached list extremes stop search, selecting none.
 *
 * If selection is none, start search from list begin (if 'local' is AFTER) or
 * start from end (if 'local' is BEFORE).
 *
 * Caution: (important note about struct pointers)
 * if 'data' parameter is a pointer to any other data structure
 * with more pointers, is important fill the fields used in nodes comparison
 * with usefull data and set to null (00) the other fields.
 */
void  class_list::
find_unsorted (relative_location local, DATAPT data)
{
	class_node *start;

	if (m_selected) {
		if (local == AFTER)
			start = m_selected->next;
		else
			start = m_selected->previous;
	}
	else {
		if (local == AFTER)
			start = m_first;
		else
			start = m_last;
	}

	m_selected = make_linear_search (AFTER, start, data);
}



/*
 * Set the compare function (function used to compare two given nodes when
 * sorting the list).
 *
 * If receive null (0) the default compare function is set.
 *
 * function - point the function to use in comparison, if NULL the default
 *            function is set. The compare functionm must return:
 *             < 0 if node1 is less than node2;
 *             ==0 if nodes are equal;
 *             > 0 if node1 is greater than node2
 *            to sort list ascending.
 */
void  class_list::
set_compare_function (DATAPT (*function) (DATAPT, DATAPT))
{
	if (function == 00)
		compare_function = default_compare_function;
	else
		compare_function = function;
}



/*
 * Revert list nodes: make first node be the last and so on.
 */
void  class_list::
revert ()
{
	// scan list
	class_node *node = m_first;
	for (DWORD index = 0; index < m_total_nodes; index++) {

		// swap node pointers
		class_node *tmp = node->next;
		node->next = node->previous;
		node->previous = tmp;

		// next node
		node = tmp;
	}

	// swap first, last pointers
	class_node *tmp = m_first;
	m_first = m_last;
	m_last = tmp;
}



/*
 * Return the node in the given position or 00 if node not found.
 */
inline class_node *class_list::
get_node_at_position (DWORD index)
{
	// verify if index is out of range, if yes return
	if (index >= m_total_nodes)
		return 00;

	// point node to return
	class_node *node;

	// verify if direct access table exist
	if (m_direct_access != 00)
		node = m_direct_access [index];
	else {

		// if index refer to last node
		if (index == m_total_nodes - 1)
			return m_last;

		// if index refer to first node
		if (index == 0)
			return m_first;

		// navigate from begin to the specified position
		node = m_first;
		for (DWORD pos = 0; pos < index; pos++)
			// verify if was found the end of list
			if (node->next == 00)
				return 00;
			else
				node = node->next;
	}

	return node;
}



/*
 * Insert the given node (new_node) before or after the
 * the node refered by 'ref_node'. The parameter 'local' define if insertion
 * would be after or before ref_node; respectively, possible values are:
 * AFTER, BEFORE.
 * Warning: parameters are not tested (designed as an internal function)
 */
inline void class_list::
insert_node (class_node * new_node,
	     relative_location local,
	     class_node * ref_node)
{
	// if list is empty (or selection is empty), insert as first node
	if (ref_node == 00) {
		m_first = m_last = m_selected = new_node;
		new_node->next = new_node->previous = 00;
		m_total_nodes++;
		return;
	}

	// insertion before reference node
	if (local == BEFORE) {

		// if insertion before first node (or if list has only one node)
		if (m_first == ref_node) {

			new_node->previous = m_first->previous;
			new_node->next = m_first;
			m_first->previous = new_node;
			m_first = new_node;
		}
		else {
			// if insertion is before the last position
			// or in middle
			ref_node->previous->next = new_node;
			new_node->previous = ref_node->previous;
			new_node->next = ref_node;
			ref_node->previous = new_node;
		}
	}
	else {
		// insertion after current node
		// if insertion after last node (or if list has only one node)
		if (ref_node == m_last) {

			new_node->next = m_last->next;
			new_node->previous = m_last;
			m_last->next = new_node;
			m_last = new_node;
		}
		else {
			// if insertion is in first or middle position
			ref_node->next->previous = new_node;
			new_node->next = ref_node->next;
			new_node->previous = ref_node;
			ref_node->next = new_node;
		}
	}

	m_total_nodes++;
}



/*
 * Remove the specified node from list, updating list pointers.
 * Don't change removed node internal pointers.
 * Warning: parameters are not tested (designed as an internal function)
 */
inline void class_list::
remove_node (class_node * node)
{
	m_total_nodes--;

	// verify if list has only one node
	if (m_first == m_last) {
		m_first = m_last = m_selected = 00;
		return;
	}

	// if node to remove is the first
	if (m_first == node) {

		// set the first as next node
		m_first->next->previous = m_first->previous;
		m_first = m_first->next;

		// if selected node is the node to be removed
		if (node == m_selected)
			// selected node continue being the first
			m_selected = m_first;

		return;
	}

	// if node to remove is the last
	if (m_last == node) {

		// set last with previous node
		m_last->previous->next = m_last->next;
		m_last = m_last->previous;

		// if selected node is the last
		if (node == m_selected)
			// select the last one
			m_selected = m_last;

		return;
	}

	// if selected node is in list middle

	// change pointers
	node->previous->next = node->next;
	node->next->previous = node->previous;

	// if selected node is the removed one
	if (node == m_selected)
		// select the previous, with same index
		m_selected = node->previous;

}



/*
 * Append source list at this list.
 * The nodes for the given list are joined after or before the given node.
 * After or before relative positions are defined in parameter 'local',
 * possible values are: AFTER or BEFORE, respectively.
 * The source list is terminateed: poiners set to 00 and source direct access freed.
 * Some members of this list are not updated: compare function pointer and
 * direct access.
 * 
 * Warning: parameters are not tested (designed as an internal function)
 */
inline void class_list::
append_list (class_list & source, relative_location local, class_node * node)
{
	// supose list empty, set variables with source list values
	if (node == 00) {

		m_first = source.m_first;
		m_last = source.m_last;
		m_selected = source.m_selected;
		m_total_nodes = source.m_total_nodes;

		// terminate source list
		source.disable_direct_access ();
		source.init ();

		return;
	}

	// if source list is empty
	if (source.m_total_nodes == 0)
		return;

	// lists has some nodes, make append

	// resolve pointers
	class_node *source_first = source.m_first;
	class_node *source_last = source.m_last;
	if (local == BEFORE) {
		// if insertion before first node (or if list has only one node)
		if (m_first == node) {

			source_first->previous = m_first->previous;
			source_last->next = m_first;
			m_first->previous = source_last;
			m_first = source_first;
		}
		else {
			// if insertion is before the last position
			// or in middle
			node->previous->next = source_first;
			source_first->previous = node->previous;
			source_last->next = node;
			node->previous = source_last;
		}
	}
	else {
		// insertion after current node
		// if insertion after last node (or if list has only one node)
		if (node == m_last) {

			source_last->next = m_last->next;
			source_first->previous = m_last;
			m_last->next = source_first;
			m_last = source_last;
		}
		else {
			// if insertion is in first or middle position
			node->next->previous = source_first;
			source_last->next = node->next;
			source_first->previous = node;
			node->next = source_first;
		}
	}

	m_total_nodes += source.m_total_nodes;

	// terminate source list
	source.disable_direct_access ();
	source.init ();
}



/*
 * Make binary tree search (see class_list::find_sorted for details).
 *
 * If found the requested data, its node is returned. If not found
 * the exactly requested data, return the most closest node.
 *
 * data - what look for
 * local - if found many match nodes select the first (using BEFORE)
 *         or the last (using AFTER);
 *         if node not found, return the closest node before (using BEFORE)
 *         or after (using AFTER).
 *
 * return:
 * total_found - number of nodes found with equal data
 * pointer to node found or 00 if impossible return any node (list empty)
 *
 * (function designed for internal use only)
 */
inline class_node *class_list::
make_btree_search (relative_location local, DATAPT data, DWORD & total_found)
{
	total_found = 0;

	// verify if list is empty, if yes return
	if (m_total_nodes == 0)
		return 00;

	DWORD ini_pos = 0;	// position of first item in segment

	DWORD end_pos = m_total_nodes - 1;	// position of last in segment

	while (ini_pos <= end_pos) {

		// get segment middle node
		// (>>1 is equal to /2, sometimes faster)
		DWORD test_pos = (ini_pos + end_pos) >> 1;
		class_node *test_node = get_node_at_position (test_pos);
		DATAPT data2 = get_data (test_node);

		// compare data
		DATAPT comp_result = (*compare_function) (data, data2);

		// test result
		if (comp_result == 0) {		// one node found

			class_node *tmp_node = test_node;

			// look for the first/last node with same comparison result
			// counting nodes
			for (;;) {

				total_found++;

				// end of list, first found
				if ((local == AFTER && test_node->next == 00)
				    || (local == BEFORE && test_node->previous == 00))
					break;

				// match with previous/next node data
				if (local == BEFORE)
					data2 = get_data (test_node->previous);
				else
					data2 = get_data (test_node->next);

				comp_result = (*compare_function) (data, data2);

				if (comp_result != 0)
					// distinct, first node was found
					break;

				// try previous/next node
				if (local == BEFORE)
					test_node = test_node->previous;
				else
					test_node = test_node->next;

			}

			// count in other direction
			local = (local == BEFORE) ? AFTER : BEFORE;	// change direction

			for (;;) {

				total_found++;

				// end of list, first found
				if ((local == AFTER && tmp_node->next == 00)
				    || (local == BEFORE && tmp_node->previous == 00))
					break;

				// match with previous/next node data
				if (local == BEFORE)
					data2 = get_data (tmp_node->previous);
				else
					data2 = get_data (tmp_node->next);

				comp_result = (*compare_function) (data, data2);

				if (comp_result != 0)
					// distinct, first node was found
					break;

				// try previous/next node
				if (local == BEFORE)
					tmp_node = tmp_node->previous;
				else
					tmp_node = tmp_node->next;

			}

			total_found--;	// must decrease one (first counted 2 times)

			// first node found (probably the first of a sequence of nodes
			// with equal comparison results)
			return test_node;
		}
		else {

			// if start and end of segment are the same, abort
			if (ini_pos == end_pos) {

				// return closest node

				// return next node, if after, if it's not null and
				// if don't refer first node (in circular lists)
				if (local == AFTER && test_node->next
				    && test_node->next != m_first)
					return test_node->next;

				return test_node;
			}
			else {
				// segment is splited in the middle up or down
				// acording result
				if (comp_result < 0)
					// select down, but makes it not negative
					end_pos = (test_pos==0) ? 0 : (test_pos-1);
				else
					// select up
					ini_pos = test_pos + 1;
			}
		}
	}

	// not found
	return 00;
}



/*
 * Make linear search from the given node to list extreme.
 * (see class_list::find for details)
 *
 * local - if BEFORE, search from 'node' to list begin; if AFTER, from 'node'
 *         to list end.
 * node - start search node
 * data - data to look for
 *
 * return:
 * pointer to node with match data or null (00) if node not found
 *
 * (function designed for internal use only)
 */
inline class_node *class_list::
make_linear_search (relative_location local, class_node * node, DATAPT data)
{
	if (!node || !m_first || !m_last)
		return 00;

	// find node
	while ((local == AFTER && node != m_last->next)		// list end
	        || (local == BEFORE && node != m_first->previous)) {	// list begin

		DATAPT data2 = get_data (node);

		// compare data
		DATAPT comp_result = (*compare_function) (data, data2);

		// test result
		if (comp_result == 0) {		// one node found

			return node;
		}

		// next
		if (local == AFTER)
			node = node->next;
		else
			node = node->previous;
	}

	// list extreme, no more found
	return 00;
}



/*
 * Default compare function used if user don't specify another.
 * (see class_list::set_compare_function, for details).
 */
DATAPT
default_compare_function (DATAPT data1, DATAPT data2)
{
	return data1 - data2;
}
