/*
 * Implement a discrete function with double words (32bit unsigned integers)
 * as domain (ids) and data pointers as image (values).
 *
 * Each id and value in map is known as an element.
 *
 * Are possible many elements with same key. In this case only the
 * last inserted would be visible (those implementation follow the stack
 * concept).
 *
 * LICENSE:
 *
 * Copyright (C) 2000 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <stdlib.h>
#include "../../include/base/type.h"
#include "../../include/base/list.h"
#include "../../include/base/mapdw.h"



struct map_element {
	DWORD id;
	DATAPT value;
};



/*
 * Compare function for map elements.
 */
DATAPT
id_comparator (DATAPT m1, DATAPT m2)
{
	map_element *el1 = (map_element *) m1;
	map_element *el2 = (map_element *) m2;

	return (DATAPT) ( int (el1->id) - int (el2->id) );
}



/*
 * constructor - init class variables
 */
class_mapdw::class_mapdw ()
{
	init ();
}



/*
 * destructor
 */
class_mapdw::~class_mapdw ()
{
	terminate ();
}



/*
 * set default values
 */
void  class_mapdw::
init ()
{
      class_list::init ();
	set_compare_function (id_comparator);
}



/*
 * Empty map (don't deallocate the values pointed areas).
 */
void  class_mapdw::
terminate ()
{
	remove_all (TRUE);
}



/*
 * Search for a given id on map.
 * return:
 * value - value related with the first id found
 * total of match elements (elements with same ids)
 */
DWORD class_mapdw::
get_value (DWORD id, DATAPT & value)
{
	map_element el;
	el.id = id;

	select_none ();
	DWORD num = find_sorted (BEFORE, get_from & el);
	if (!num)
		return 0;

	map_element *ref;
	get_selected (put_in ref);
	value = ref->value;
	return num;
}



/*
 * Get element in the given position.
 * return:
 * value - the element value
 * the id or -1 if position was not found
 */
DWORD class_mapdw::
get_at_position (DWORD index, DATAPT & value)
{
	map_element *ref;
	BOOL  success = class_list::get_at_position (index, put_in ref);

	if (success) {
		value = ref->value;
		return ref->id;
	}

	return (DWORD) -1;
}



/*
 * Add a new element to map.
 * return:
 * TRUE - ok / FALSE - error
 */
BOOL  class_mapdw::
insert (DWORD id, DATAPT value)
{
	map_element *ref = (map_element *) malloc (sizeof (map_element));
	if (!ref)
		return FALSE;

	ref->id = id;
	ref->value = value;

	// insert
	BOOL  success = insert_first (BEFORE, get_from ref);
	if (!success) {
		free (ref);
		return FALSE;
	}

	enable_direct_access ();

	sort ();

	return TRUE;
}



/*
 * Removes the first element found with the given id from map.
 * return:
 * value - the element value
 * the number of remaining elements with this same id or -1 if none.
 */
DWORD class_mapdw::
remove (DWORD id, DATAPT & value)
{
	map_element el;
	el.id = id;
	el.value = 0;

	select_none ();
	DWORD num = find_sorted (BEFORE, get_from & el);
	if (!num)
		return 0;

	map_element *ref;
	remove_selected (put_in ref);
	value = ref->value;

	free ( ref );

	enable_direct_access ();

	return num - 1;
}
