/*
 * Implement a discrete function with strings as domain (keys) and
 * pointers as image (values).
 *
 * Each key and value in map is known as an element.
 *
 * Are possible many elements with same key. In this case only the
 * last inserted would be visible (those implementation follow the stack
 * concept).
 *
 * LICENSE:
 *
 * Copyright (C) 2000 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <stdlib.h>
#include "../../include/base/type.h"
#include "../../include/base/list.h"
#include "../../include/base/mapstr.h"



struct map_element {
	char *key;
	DATAPT value;
};



/*
 * Compare function for map elements.
 */
DATAPT
element_comparator (DATAPT data1, DATAPT data2)
{
	map_element *el1 = (map_element *) data1;
	map_element *el2 = (map_element *) data2;

	return (DATAPT) strcmp (el1->key, el2->key);
}



/*
 * constructor - init class variables
 */
class_mapstr::class_mapstr ()
{
	init ();
}



/*
 * destructor
 */
class_mapstr::~class_mapstr ()
{
	terminate ();
}



/*
 * set default values
 */
void  class_mapstr::
init ()
{
      class_list::init ();
	set_compare_function (element_comparator);
}



/*
 * Empty map (don't deallocate the values and keys pointed areas).
 */
void  class_mapstr::
terminate ()
{
	remove_all (TRUE);
}



/*
 * Search for a given key on map.
 * return:
 * value - value related with the first key found
 * total of match elements (elements with same key)
 */
DWORD class_mapstr::
get_value (char *key, DATAPT & value)
{
	map_element el;
	el.key = key;
	el.value = 0;

	select_none ();
	DWORD num = find_sorted (BEFORE, get_from & el);
	if (!num)
		return 0;

	map_element *ref;
	get_selected (put_in ref);
	value = ref->value;
	return num;
}



/*
 * Get element in the given position.
 * return:
 * value - the element value
 * the key string or 00 if position was not found
 */
char *class_mapstr::
get_at_position (DWORD index, DATAPT & value)
{
	map_element *ref;
	BOOL  success = class_list::get_at_position (index, put_in ref);

	if (success) {
		value = ref->value;
		return ref->key;
	}

	return 00;
}



/*
 * Add a new element to map.
 * key - element key
 * value - element value
 *
 * Key string is not duplicated, don't deallocate it until
 * this class instance is in use.
 *
 * return:
 * TRUE - ok / FALSE - error
 */
BOOL  class_mapstr::
insert (char *key, DATAPT value)
{
	map_element *ref = (map_element *) malloc (sizeof (map_element));
	if (!ref)
		return FALSE;

	ref->key = key;
	ref->value = value;

	// try to insert element keeping map sorted
	//find_sorted (BEFORE, get_from ref); (resort if select nothing)

	// insert
	BOOL  success = insert_first (BEFORE, get_from ref);
	if (!success) {
		free (ref);
		return FALSE;
	}

	enable_direct_access ();

	sort ();

	return TRUE;
}



/*
 * Removes the first element found with the given key from map.
 * Don't delete the key or the value.
 * return:
 * value - the element value
 * pointer to key in map, or 00 if no key found
 */
char* class_mapstr::
remove (char *key, DATAPT & value)
{
	map_element el;
	el.key = key;
	el.value = 0;

	select_none ();
	DWORD num = find_sorted (BEFORE, get_from & el);
	if (!num)
		return 0;

	map_element *ref;
	remove_selected (put_in ref);
	value = ref->value;
	char* map_key = ref->key;

	free (ref);

	enable_direct_access ();

	return map_key;
}
