/*
 * Base Lao list component.
 * 
 * LICENSE:
 *
 * Copyright (C) 2000 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <stdio.h>
#include "../../include/sbase.h"
#include "../../include/lao/colist.h"



class_colist::class_colist ()
{
  init ();
}



class_colist::~class_colist ()
{
}



void  class_colist::
init ()
{
  class_list::init ();
}



/*
 * Remove all elements from the list.
 */
void  class_colist::
terminate (class_enviroment * env, class_sbase_history* hist)
{
  class_component *rm_node;
  while (remove_first (put_in rm_node)) {
    rm_node->call (BIN_CLEAN_DATA, env, hist, 0);
    rm_node->terminate ();
    free (rm_node);
  }

  class_list::terminate ();
}



/*
 * Add the given component to list.
 * comp - new component
 * 
 * return:
 * TRUE - ok / FALSE - fail when creating or adding
 * Update history acording the success or failure of the operation.
 */
BOOL  class_colist::
add_component (class_component* comp, class_enviroment* env,
	       class_sbase_history* hist)
{
  if (!comp) return TRUE;  // REVIEW: history warning?

  BOOL  success = insert_last (AFTER, (DATAPT) comp);
  if (success) return TRUE;

  // in case of error
  comp->terminate ();
  free (comp);
  hist->add_predef_text (LH_TEXT, "error2");
  hist->add_flag (LH_ERROR, ERR_NO_MEMORY);
  hist->set_exception (LE_EXIT | LE_ERROR | LE_FATAL_ERROR);
  return FALSE;
}



/*
 * Return the textual representation of the list.
 * open_ch - the open list symbol (char)
 * close_ch - the close list symbol (char)
 * env - enviroment
 * hist - history
 * return:
 * class_string pointer to string which represents the list
 */
DATAPT class_colist::
get_text_representation ( char open_ch, char close_ch, class_enviroment * env, class_sbase_history* hist )
{
  // begin list textual representation
  class_string *str = (class_string *) malloc (sizeof (class_string));
  str->init ();
                        
  str->append_char (open_ch);
  str->append_char (' ');

  // for each component in list
  class_component *comp;
  select_first ();
  while (get_selected (put_in comp)) {

    // get text from component
    class_string *comp_str = (class_string *) comp->call (BIN_GET_TEXT_REPRESENTATION, env, hist, 0);

    str->append (comp_str->get ());
    str->append_char (' ');

    comp_str->terminate ();
    free (comp_str);
    select_next ();
  }

  str->append_char (close_ch);
  return (DATAPT) str;
}



/*
 * Function to handle messages to list.
 * self - class_component instance for this object
 * msg - lao system message ID
 * env - reference to lao name table (enviroment)
 * hist - reference to history
 * pdata - points component data
 * param - parameter to component
 *
 * return:
 * message process result
 */
DATAPT
call_list (class_component * self, DATAPT msg, class_enviroment * env,
           class_sbase_history* hist, DATAPT * pdata, DATAPT param)
{
  DATA_POINTED (pdata, hist);

  switch (get_sysmsg (msg)) {

  case BIN_CREATE_DATA:
    {
      DATA_UNAVAILABLE (*pdata, hist);
      
      // set component data member:
      // expect: param as a pointer to class_colist instance,
      // or one derived class from class_colist
      *pdata = param;
      return *pdata;
    }
    break;

  case BIN_CLEAN_DATA:
    {
      DATA_AVAILABLE (*pdata, hist);
      
      class_colist *lst = (class_colist *) * pdata;

      lst->terminate (env, hist);

      delete pdata; // REVIEW ?
      *pdata = 0;
      return *pdata;
    }
    break;

  case INVOKE:
  case INVOKE_AS_REFERENCE:
  case ACTIVATE:
  case ACTIVATE_AS_REFERENCE:
  case ACTIVATE_AS_PARAMETER:
    {
      DATA_AVAILABLE (*pdata, hist);

      // activate the list
      class_colist *lst = (class_colist *) * pdata;
      return lst->activate ( self, msg, env, hist, param);
    }
    break;

  case BIN_ACTIVATE:
    {
      DATA_AVAILABLE (*pdata, hist);

      class_colist *lst = (class_colist *) * pdata;

      // execute one specific service
      switch (get_actsrv (msg)) {

      case GET_ELEMENT:
	// return the list element at the position defined in param
	{
	  DATAPT element;
	  BOOL res = lst->get_at_position (param, element);
	  if (res)
	    return element;
	  else
	    return 00;
	}
	break;

      case NUMBER_OF_ELEMENTS:
	// return the number of elements in list
	return lst->get_total_nodes ();
	break;
      }
    }
    break;

  case BIN_ADD_COMPONENT:
    {
      DATA_AVAILABLE (*pdata, hist);
      if (get_actsrv (msg) == 0 ||
          get_actsrv (msg) == 3 ) {   // if recognized one common token
	// add the given component to list
	class_colist *lst = (class_colist *) * pdata;
	return lst->add_component ( (class_component*) param, env, hist);
      }
      return TRUE;
    }
    break;

  case BIN_PROCESS_SYMBOL:
  case BIN_SYNC_RECOGNITION:
  case BIN_RECOGNITION_TOKEN_BEGIN:
  case BIN_RECOGNITION_TOKEN_SUCC:
  case BIN_RECOGNITION_FAIL:
  case BIN_RETURN_COMPONENT:
    {
      return 00;
    }
    break;

  case BIN_GET_TEXT_REPRESENTATION:
    {
        DATA_AVAILABLE (*pdata, hist);

	class_colist *lst = (class_colist *) * pdata;
	return lst->get_text_representation (env, hist);
    }
    break;

  case COPY:
    return (DATAPT) self;
    break;
  }
  
  return 00;
}
