/*
 * String component.
 * Handle standard strings and components name manipulation.
 * 
 * LICENSE:
 *
 * Copyright (C) 2000 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <stdio.h>
#include "../../include/sbase.h"
#include "../../include/lao/costring.h"



// define operators flags
#define op_INSERT 0
#define op_REMOVE 1
#define LAST_ID 2



// the operators table - makes relation between
// operator ID and operator string
static l_op_element OPS_TAB [] = {
  { op_INSERT, "=" },
  { op_REMOVE, "remove" },
};



class_costring::class_costring ()
{
  init ();
}



class_costring::~class_costring ()
{
  terminate ();
}



void  class_costring::
init ()
{
  class_string::init ();
}



/*
 * Empty string.
 */
void  class_costring::
terminate ()
{
  class_string::terminate ();
}



/*
 * Process the parameters given to the string component
 * by the list.
 * self - this component
 * env - enviroment
 * hist - history
 * param - activation parameters (the list of components)
 * return:
 * the activation result
 */
DATAPT class_costring::
process_parameters (class_component* self, class_enviroment* env,
                    class_sbase_history* hist, DATAPT param)
{
  class_component *ret_comp = self;

  // parameters list available
  if (param) {

    // scan parameters list 
    DWORD total_params = paramc (param, env, hist);
    int state = 0;
    // for each parameter (component), expect the first one (self)
    for ( DWORD par_pos = 1; par_pos < total_params; par_pos ++ ) {

      // get the parameter
      class_component *par_comp = paramv (par_pos, param, env, hist);

      // activate the parameter
      class_component *res_comp = (class_component *) par_comp->call (ACTIVATE_AS_PARAMETER, env, hist, param);

      // what is the returned component
      switch (state) {

      // verify if component is an operator flag
      case 0:
        {
          // define the text representation of the returned component
          class_string *res_str = (class_string *) res_comp->call (BIN_GET_TEXT_REPRESENTATION, env, hist, 0);

          l_op_id op_id = self->parse_param_str ( res_str->get (), LAST_ID, OPS_TAB );

          // what operator the parameter is
          switch (op_id) {
          case op_INSERT:
            state = 1;
            break;

          case op_REMOVE:
            {
              // component as a name 
              // remove name from enviroment
              class_component* value;
              env->remove (get (), &value);
              //value->call ( BIN_CLEAN_DATA, env, 0);
              //value->terminate ();
              ret_comp = value;
            }
            break;

          default:
            //(REVIEW: what to do if operator not recognized)
            break;
          }
        }
        break;

      // first component as a name
      case 1:
        // insert in enviroment the name-value (attribution)
        env->insert (get (), res_comp);
        ret_comp = res_comp;
        state = 0;
        break;

      } // parse components
    }   // for each parameter
  }     // if receive parameters
  else
    ret_comp = self;

  return (DATAPT) ret_comp;
}



/*
 * Function to handle component messages.
 * self - class_component instance for this object
 * msg - lao system message ID
 * env - reference to lao name table
 * hist - reference to history
 * pdata - points component data
 * param - component parameter
 *
 * return:
 * process result
 */
DATAPT
call_string (class_component * self, DATAPT msg, class_enviroment * env,
             class_sbase_history* hist, DATAPT * pdata, DATAPT param)
{

  DATA_POINTED (pdata, hist);

  switch (get_sysmsg (msg)) {

  case BIN_CREATE_DATA:
    {
      DATA_UNAVAILABLE (*pdata, hist);
      
      *pdata = (DATAPT) malloc (sizeof (class_costring));
      MEMORY_ALLOCATED (*pdata, hist);

      // define information
      class_costring *costr = (class_costring *) * pdata;
      costr->init ();

      // expect: param as char*
      costr->append ((char *) param);

      return *pdata;
    }
    break;

  case BIN_CLEAN_DATA:
    {

      DATA_AVAILABLE (*pdata, hist);

      class_costring *costr = (class_costring *) * pdata;
      costr->terminate ();

      free ((void *) *pdata);
      *pdata = 0;
      return *pdata;
    }
    break;

  case INVOKE:
    {
      class_component *discardable_comp = (class_component *) ::call_string (self, COPY, env, hist, pdata, 0);
      if (!discardable_comp) return 00;
      return discardable_comp->call (ACTIVATE, env, hist, param);
    }
    break;

  case INVOKE_AS_REFERENCE:
    {
      return ::call_string (self, ACTIVATE_AS_REFERENCE, env, hist, pdata, param);
    }
    break;

  case ACTIVATE:
    {
      class_costring *costr = (class_costring *) * pdata;
      return costr->process_parameters (self, env, hist, param);
    }
    break;

  case ACTIVATE_AS_REFERENCE:
  case ACTIVATE_AS_PARAMETER:
    return (DATAPT) self;
    break;

  case BIN_ACTIVATE:
  case BIN_PROCESS_SYMBOL:
  case BIN_SYNC_RECOGNITION:
  case BIN_RECOGNITION_TOKEN_BEGIN:
  case BIN_RECOGNITION_TOKEN_SUCC:
  case BIN_RECOGNITION_FAIL:
  case BIN_RETURN_COMPONENT:
  case BIN_ADD_COMPONENT:
    {
      return 00;
    }
    break;
    
  case BIN_GET_TEXT_REPRESENTATION:
    {
      class_string *str = (class_string *) malloc (sizeof (class_string));
      str->init ();
      class_costring *comp_str = (class_costring *) * pdata;
      str->append (comp_str->get ());
      return (DATAPT) str;
    }
    break;

  case COPY:
    {
      DATA_AVAILABLE (*pdata, hist);

      class_costring *costr = (class_costring *) * pdata;

      return (DATAPT) build_component ((DATAPT) costr->get (), env, hist, ::call_string, "<string>", "<string>");
    }
    break;
  }

  return 00;
}
