/*
 * Implements a component to parse symbols of the type character.
 * 
 * LICENSE:
 *
 * Copyright (C) 2000 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <stdio.h>
#include "../../include/sbase.h"
#include "../../include/lao/laodefs.h"
#include "../../include/lao/cochar.h"
#include "../../include/lao/pchar.h"



class_pchar::class_pchar ()
{
  init ();
}



class_pchar::~class_pchar ()
{
  terminate ();
}



void  class_pchar::
init ()
{
  m_str.init ();
  m_state = 0;
}



void  class_pchar::
terminate ()
{
  m_str.terminate ();
}



/*
 * Makes the state machine process the given symbol.
 * sym - symbol to process
 * return:
 * machine result
 */
DATAPT  class_pchar::
process_symbol (symbol_info* sym)
{
  class_char* cho = &(sym->obj);
  char ch = cho->get ();
  l_sysmsg ret_val = NONE;

  switch (m_state) {

  case 0: 	// start recognition
    if (sym->ct->match (ch, CHC_CHAR_DELIMITER)) {   // begin
      ret_val = BIN_RECOGNITION_TOKEN_BEGIN;
      m_state = 1;
      m_str.terminate ();
      m_str.init ();
    } else {
      if (! sym->ct->match (ch, CHC_SEPARATORS)) {  // invalid char input
	ret_val = BIN_RECOGNITION_FAIL;
	m_state = 3;
      }
    }
    // wait next input
    break;

  case 1:   // process char
    if (sym->ct->match (ch, CHC_CONTROLS)) {       // invalid chars
      ret_val = BIN_RECOGNITION_FAIL;
      m_state = 3;
    } else {                  // continue
      m_str.append_char (cho->get ());
      m_state = 2;
    }
    break;
    
  case 2:
    if (sym->ct->match (ch, CHC_CHAR_DELIMITER)) {   // end, success
      ret_val = BIN_RECOGNITION_TOKEN_SUCC;
    }	else {                                     // end, fail
      ret_val = BIN_RECOGNITION_FAIL;
    }
    m_state = 3;
    break;
    
  case 3:   // char recognition terminated
    break;
  }

  return (DATAPT) ret_val;
}



/*
 * Return the component which represents the recognized
 * token or structure.
 * return:
 * pointer to the returned component or data
 */
DATAPT  class_pchar::
return_component ( class_enviroment* env, class_sbase_history* hist )
{
  return (DATAPT) ::build_component ((DATAPT) m_str.get (), env, hist, ::call_char, "<pchar>", "<char>");
}



/*
 * Reset the parse state machine to the initial state.
 * Preparing to begin a token recognition.
 * return:
 * NULL
 */
DATAPT  class_pchar::
sync_recognition ( class_enviroment* env, class_sbase_history* hist )
{
  m_state = 0;
  return 00;
}



/*
 * Function to handle component messages.
 * self - class_component instance for this object
 * msg - lao system message ID
 * env - reference to lao name table
 * hist - reference to history
 * pdata - points component data
 * param - component parameter
 *
 * return:
 * process result
 */
DATAPT
call_pchar (class_component * self, DATAPT msg, class_enviroment * env,
             class_sbase_history* hist, DATAPT * pdata, DATAPT param)
{
  DATA_POINTED (pdata, hist);

  switch (get_sysmsg (msg)) {

  case BIN_CREATE_DATA:
    {
      DATA_UNAVAILABLE (*pdata, hist);

      *pdata = (DATAPT) malloc (sizeof (class_pchar));
      MEMORY_ALLOCATED (*pdata, hist);

      // define information
      class_pchar *och = (class_pchar *) * pdata;
      och->init ();

      return *pdata;
    }
    break;

  case BIN_CLEAN_DATA:
    {
      DATA_AVAILABLE (*pdata, hist);

      class_pchar *och = (class_pchar *) * pdata;
      och->terminate ();

      free ((void *) *pdata);
      *pdata = 0;
      return *pdata;
    }
    break;

  case BIN_PROCESS_SYMBOL:
    {
      DATA_AVAILABLE (*pdata, hist);

      symbol_info* sym = (symbol_info*) param;
		
      class_pchar *och = (class_pchar *) * pdata;
      return och->process_symbol (sym);
    }
    break;	  

  case BIN_RETURN_COMPONENT:
    {
      DATA_AVAILABLE (*pdata, hist);

      class_pchar *och = (class_pchar *) * pdata;
      return och->return_component (env, hist);
    }
    break;

  case BIN_SYNC_RECOGNITION:
    {
      DATA_AVAILABLE (*pdata, hist);

      class_pchar *och = (class_pchar *) * pdata;
      return och->sync_recognition (env, hist);
    }
    break;

  case BIN_RECOGNITION_TOKEN_BEGIN:
  case BIN_RECOGNITION_TOKEN_SUCC:
  case BIN_RECOGNITION_FAIL:
  case BIN_ADD_COMPONENT:
  case BIN_ACTIVATE:
  case INVOKE:
  case ACTIVATE:
  case ACTIVATE_AS_REFERENCE:
  case ACTIVATE_AS_PARAMETER:
    {
       return 00;
    }
    break;

  case BIN_GET_TEXT_REPRESENTATION:
    {
      class_string *str = (class_string *) malloc (sizeof (class_string));
      str->init ();
      str->append ("class_pchar instance");
      return (DATAPT) str;
    }
    break;

  case COPY:
    {
        return (DATAPT) build_component ((DATAPT) NULL, env, hist, ::call_pchar, "<pchar>", "<pchar>");
    }
    break;
  }

  return 0;
}
