/*
 * Arrange and manage the parser components in a way they can
 * recognize tokens. The tokens are classified per type. Examples of types
 * are: character, integer, strings, lists, arrays, etc. Each type is
 * recognized by one parser.
 * For example: the type char is defined by the sequence of symbols: '<char>',
 * where <char> can be any character. To recognize correctly any
 * token, each symbol read is sent to parsers of all types. And only the
 * parser of type char (in this case) will recognize, the others will fail
 * at some symbol.
 * Each type has a stack with one or more parsers. The parser in stack top
 * receive symbols. If the parser in stack top recognize a token, return it.
 * Sometimes parsers can be poped or pushed from stack. If one structure
 * begin is recognized, then a parser of that type is pushed. When the
 * structure end is recognized, then it is poped. For each type exists,
 * at least, one parser in stack.
 *
 * LICENSE:
 *
 * Copyright (C) 2000 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include "../../include/base.h"
#include "../../include/sbase/def.h"
#include "../../include/sbase/shistory.h"
#include "../../include/sbase/env.h"
#include "../../include/sbase/comp.h"
#include "../../include/sbase/mparser.h"



class_multiparser::class_multiparser ()
{
  init ();
}



class_multiparser::~class_multiparser ()
{
}



void  class_multiparser::
init ()
{
  m_parsers.init ();
  m_parse_pos = 0;
  m_num_parsers_type = 0;
  m_all_pos = 0;
}



/*
 * Terminate all parsers.
 */
void  class_multiparser::
terminate (class_enviroment* env, class_sbase_history* hist)
{
  class_component *rm_parser;
  DWORD pos = 0;
  while (m_parsers.get_at_position (pos, put_in rm_parser) != 00) {
    rm_parser->call (BIN_CLEAN_DATA, env, hist, 0);
    rm_parser->terminate ();
    free (rm_parser);
    pos ++;
  }

  m_parsers.terminate ();
  init ();
}



/*
 * Define one new parser.
 * pname - name of the type to be recognized with this parser
 * comp - parser that recognize the tokens of this type
 * return:
 * TRUE - success; FALSE - error: no memory, or name still defined
 */
BOOL class_multiparser::
add_type (char* pname, class_component* comp)
{
  DATAPT value;
  DWORD num;
  num = m_parsers.get_value (pname, put_in value);
  if (num == 0)
    return m_parsers.insert (pname, get_from comp);

  return FALSE;
}



/*
 * Get the parser component for the type currently in use.
 * return:
 * the component or 00 if none
 */
class_component*  class_multiparser::
get_parser_in_use ()
{
  class_component *comp;
  char* pname;
  pname = m_parsers.get_at_position (m_parse_pos, put_in comp);
  m_parsers.get_value (pname, put_in comp);

  return comp;
}



/*
 * Make a copy of the parser in use, and place the copy as the current
 * parser for this type. It is usefull to recognize structures
 * recursively.
 * env - sbase enviroment
 * hist - script history
 * return:
 * TRUE - ok; FALSE - error
 */
BOOL class_multiparser::
push_parser (class_enviroment* env, class_sbase_history* hist)
{
  // get the component to be used as source
  class_component *comp;
  char* pname;
  pname = m_parsers.get_at_position (m_parse_pos, put_in comp);
  DWORD num;  
  num = m_parsers.get_value (pname, put_in comp);
  if (num == 0) return FALSE;

  class_component* copy_comp = (class_component*) comp->call (COPY, env, hist, 0);
  copy_comp->call (BIN_SYNC_RECOGNITION, env, hist, 0);

  return m_parsers.insert (pname, get_from copy_comp);
}



/*
 * Pop the current parser in use, if exist more than one
 * for the specified type.
 * return:
 * the component or 00 if none could be poped
 */
class_component* class_multiparser::
pop_parser ()
{
  class_component *comp;
  char* pname;
  pname = m_parsers.get_at_position (m_parse_pos, put_in comp);
  DWORD num;  
  num = m_parsers.get_value (pname, put_in comp);
  if (num <= 1) return 00;
  
  m_parsers.remove (pname, put_in comp);
  return comp;
}



/*
 * Start the activation of all top parsers.
 */
void class_multiparser::
start_activate_all ()
{
  m_all_pos = 0;
}



/*
 * Send the given message to some (or all) parsers that can be used
 * for their respective types (they are the parsers in top of their
 * respective stacks).
 * the_msg - a sbase message to send
 * env - environment
 * hist - sbase execution history
 * parameter - activation parameter
 * skip_inuse - TRUE - skip (don't send the message to) the parser
 *              which corresponds of the type in use, if FALSE - send to all
 * return:
 * 0 - if all needed components were activated; another value
 * specifying one message to translator. In this case the activation
 * for other components is interrupted, must continue later, until
 * this method return 0.
 */
DATAPT class_multiparser::
activate_all ( DATAPT the_msg, class_enviroment* env,
               class_sbase_history* hist, DATAPT param,
               BOOL skip_inuse )
{     
  // define the number of elements
  DWORD num_els = m_parsers.get_total_nodes ();

  // try activate all top parsers
  while (m_all_pos < num_els) {
    char* pname;
    class_component* comp;
    pname = m_parsers.get_at_position (m_all_pos, put_in comp);
    DWORD num;
    num = m_parsers.get_value (pname, put_in comp);
    
    // define the next parser which has a distinct name
    DWORD old_pos = m_all_pos;
    m_all_pos += num;
  
    // verify if this is the name of parser to skip
    if (old_pos != m_parse_pos || ! skip_inuse) {
      // activate, sending the message
      DATAPT res = comp->call (the_msg, env, hist, param);
      // this parser return one result: interrupt to process it
      if (res != 00) return res;
    }
  }

  return 00;
}



/*
 * Prepare to activate the parsers to recognize a symbol.
 * return:
 * TRUE - ok; FALSE - error (none in use)
 */
BOOL class_multiparser::
use_first_type ()
{
  m_parse_pos = 0;    // use the first parser
  char* pname;
  class_component* comp;
  pname = m_parsers.get_at_position (m_parse_pos, put_in comp);
  m_num_parsers_type = m_parsers.get_value (pname, put_in comp);
  if (m_num_parsers_type == 0) return FALSE;
  return TRUE;
}



/*
 * Define to use the parser for next type, if there is no more types
 * return FALSE, else TRUE.
 */
BOOL class_multiparser::
use_next_type ()
{
  DWORD num_els = m_parsers.get_total_nodes ();

  m_parse_pos += m_num_parsers_type != 0 ? m_num_parsers_type : 1;
  if (m_parse_pos >= num_els) return FALSE;

  return TRUE;
}



/*
 * Activate the parser of the type in use to recognize the given symbol.
 */
DATAPT class_multiparser::
activate_type (DATAPT the_msg, class_enviroment* env,
               class_sbase_history* hist, DATAPT param)
{
  // activate the parser
  char* pname;
  class_component* comp;
  pname = m_parsers.get_at_position (m_parse_pos, put_in comp);
  m_num_parsers_type = m_parsers.get_value (pname, put_in comp);

  // activate, sending the message
  return comp->call (the_msg, env, hist, param);
}
