/*
 * Implement the history - execution log.
 * This class derive from class_history, see it for more information.
 * 
 * LICENSE:
 *
 * Copyright (C) 2000 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <stdio.h>
#include "../../include/base.h"
#include "../../include/sbase/def.h"
#include "../../include/sbase/shistory.h"



// define messages
struct msg_element {
  char* key;
  char* msg;
};

msg_element STDMSGS [] = {
  {"error0", "none"},
  {"error1", "undefined error"},
  {"error2", "not enough memory"},
  {"error3", "name not found in enviroment"},
  {"error4", "ERR_PARSE_CONTROL"},
  {"error5", "ERR_PARSE_EOI"},
  {"error6", "ERR_PARSE_EOL"},
  {"error7", "ERR_PARSE_CHAR"},
  {"error8", "script internal error - NULL data received while activating."},
  {"error9", "component has data, can't create more."},
  {"error10", "component don't have data."},
  {"debug0", "none"},
  {"debug1", "show sintax processing"},
  {"debug2", "breakpoint set"},
  {"debug4", "trace components activation"},
  {"debug8", "show activated components text representation"},
  {"", ""},   // the end
};



class_sbase_history::class_sbase_history ()
{
    init ();
}



class_sbase_history::~class_sbase_history ()
{
}



void  class_sbase_history::
init ()
{
  class_history::init ();
  m_fatal_error = FALSE;
}



/*
 * Clear history completely.
 */
void  class_sbase_history::
terminate ()
{
  // empty history log
    BOOL end = FALSE;
    while ( !end ) {

      char* str = 00;
      char* label = get_entry ( 0, 00, TRUE, put_in str );
      
      // if has found one entry
      if ( label ) {

        // clear it
        if (strcmp (label, LH_TEXT) == 0)
          free (str);

        free (label);

        remove_selected_entry ();
      }
      else
        end = TRUE;
    }

    // empty predefined messages table
    char* key;
    DATAPT data;
    while ((key = remove_any_predefined_data (data))) {
      free (key);
      free ((void*)data);
    }

  class_history::terminate ();

  init ();
}



/* Define sbase standard messages.
   return:
   TRUE - ok / FALSE - error
 */
BOOL class_sbase_history::
predefine_messages ()
{
  char* key = 00;
  msg_element* el = STDMSGS;
  while ( *el->key != 0 ) {
    
    class_string str;
    key = str.duplicate (el->key);
    char* msg = str.duplicate (el->msg);

    BOOL res;
    res = predefine_data ( key, (DATAPT) msg );
    if (!res) {
      m_fatal_error = TRUE;
      return FALSE;
    }

    el ++;   // next register
  }
  return TRUE;
}



/*
 * Get the first error code found.
 *
 * return:
 * the error code (zero if no error)
 */
l_error class_sbase_history::
get_error ()
{
  DATAPT data;

  data = 0;
  get_entry ( 0, LH_ERROR, FALSE, data );
  return (l_error) data;
}



/*
 * Get the first debug mode found in history.
 *
 * return:
 * the debug mode (zero if label not found)
 */
l_debug class_sbase_history::
get_debug_mode ()
{
  DATAPT data = 0;

  get_entry ( 0, LH_DEBUG, FALSE, data );
  return (l_debug) data;
}



/*
 * Get the first exit code found. The entry label "exit-code"
 * is looked for.
 *
 * return:
 * the exit code, or 0 if not found
 */
int class_sbase_history::
get_exit_code ()
{
  DATAPT data;

  data = 0;
  get_entry ( 0, LH_EXIT_CODE, FALSE, data );
  return (int) data;
}



/*
 * Return TRUE if history reports some exception entry where
 * exit (stop execution) is set TRUE; else return FALSE.
 */
BOOL class_sbase_history::
has_exit ()
{
  // get selected entry
  DATAPT flag;
  char* label = get_selected_entry ( flag );

  // test if the selected entry has the right label
  if ( ! label || strcmp (label, LH_EXCEPTION) != 0 )
    // get exception looking for it in history
    label = get_entry ( 0, LH_EXCEPTION, FALSE, flag );

  return flag & LE_EXIT;
}



/*
 * Return TRUE if history reports some exception entry where
 * warning (interpretation incidents) is set TRUE; else return FALSE.
 */
BOOL class_sbase_history::
has_warning ()
{
  // get selected entry
  DATAPT flag;
  char* label = get_selected_entry ( flag );

  // test if the selected entry has the right label
  if ( ! label || strcmp (label, LH_EXCEPTION) != 0 )
    // get exception looking for it in history
    label = get_entry ( 0, LH_EXCEPTION, FALSE, flag );

  return flag & LE_WARNING;
}



/*
 * Return TRUE if history reports some exception entry where
 * debug mode is set TRUE ("debugger is on"; else return FALSE.
 */
BOOL class_sbase_history::
has_debug ()
{
  // get selected entry
  DATAPT flag;
  char* label = get_selected_entry ( flag );

  // test if the selected entry has the right label
  if ( ! label || strcmp (label, LH_EXCEPTION) != 0 )
    // get exception looking for it in history
    label = get_entry ( 0, LH_EXCEPTION, FALSE, flag );

  return flag & LE_DEBUG;
}



/*
 * Return TRUE if history reports some exception entry where
 * error presence is set TRUE; else return FALSE.
 */
BOOL class_sbase_history::
has_error ()
{
  // get selected entry
  DATAPT flag;
  char* label = get_selected_entry ( flag );

  // test if the selected entry has the right label
  if ( ! label || strcmp (label, LH_EXCEPTION) != 0 )
    // get exception looking for it in history
    label = get_entry ( 0, LH_EXCEPTION, FALSE, flag );

  return flag & LE_ERROR;
}



/*
 * Return TRUE if history reports fatal error;
 * else return FALSE.
 */
BOOL class_sbase_history::
has_fatal_error ()
{
  return m_fatal_error;
}



/* Send history output to the given stream.
   dest - stream to send data
 */
void class_sbase_history::
show (FILE* dest)
{
    // show text entries    
    BOOL end = FALSE;
    DWORD index = 0;
    while ( !end ) {

      char* str = 00;
      char* label = get_entry ( index, LH_TEXT, FALSE, put_in str );
      
      // if have found one entry
      if ( label ) {
        fprintf (dest, "%s\n", str );
        index ++;
      }
      else
        end = TRUE;
    }
}



/* Update history adding one entry with general data.

   label - the given label
   data - entry data

   Acording the label, select the entry on the history stack.

   return:
   TRUE - ok / FALSE - error when updating
 */
BOOL class_sbase_history::
add_flag ( char* label, DATAPT data)
{
  class_string new_lab;
  char* lab = new_lab.duplicate (label);
  if (! lab ) {
    m_fatal_error = TRUE;
    return FALSE;
  }

  BOOL res = update ( lab, data, FALSE );
  if (!res) m_fatal_error = TRUE;
  return res;
}



/* Update history adding one entry with text data.

   label - the given label
   str - the entry string

   return:
   TRUE - ok / FALSE - error when updating
 */
BOOL class_sbase_history::
add_text ( char* label, char* str)
{
  class_string new_str;

  char* lab = new_str.duplicate (label);
  if (! lab ) {
    m_fatal_error = TRUE;
    return FALSE;
  }

  char* ns = new_str.duplicate (str);
  if (! ns ) {
    m_fatal_error = TRUE;
    free ( lab );
    return FALSE;
  }
  
  BOOL res = update ( lab, (DATAPT) ns, FALSE );

  if ( ! res ) {
    m_fatal_error = TRUE;
    free (lab);
    free (ns);
    return FALSE;
  }

  return TRUE;
}



/* Update history adding one entry with an
   predefined text.

   label - the given label
   key - the predefined text key

   return:
   TRUE - ok / FALSE - error when updating
 */
BOOL class_sbase_history::
add_predef_text ( char* label, char* key)
{
  // get predefined text
  char* predef;
  BOOL res;
  res = get_predefined_data ( key, put_in predef );
  if ( !res ) {
    m_fatal_error = TRUE;
    return FALSE;
  }

  // alloc entry
  class_string new_str;

  char* lab = new_str.duplicate (label);
  if (! lab ) {
    m_fatal_error = TRUE;
    return FALSE;
  }

  char* str = new_str.duplicate (predef);
  if (! str ) {
    m_fatal_error = TRUE;
    free ( lab );
    return FALSE;
  }

  res = update ( lab, (DATAPT) str, FALSE );

  if ( ! res ) {
    m_fatal_error = TRUE;
    free (lab);
    free (str);
    return FALSE;
  }

  return TRUE;
}



/* Look for a predefined data and return as an text data.

   key - the predefined text key

   return:
   pointer to text or 00 if not found
 */
char* class_sbase_history::
get_predef_text ( char* key)
{
  char* predef;
  BOOL res;
  res = get_predefined_data ( key, put_in predef );
  if ( !res ) {
    m_fatal_error = TRUE;
    return 00;
  }

  return predef;
}



/* Add to history one entry with an exception update.
   Exception is one entry with information about the 
   general running status of the interpreter.

   flags_id - the given flags to set - changing status

   The last flags for the previous exception entry found
   are manteined (copied to the new one), and are set to 1 (TRUE)
   the given flags.

   The previous entry is looked for with two ways:
   - verifying if the selected entry label has the string
     expanted in macro LH_EXCEPTION, and use the data field.
   - if selected entry label is not the LH_EXCEPTION string, search
     for the first entry with this contents, and use the data field.

   return:
   TRUE - ok / FALSE - error when updating
 */
BOOL class_sbase_history::
set_exception ( DWORD flags_id )
{
  // set the especial entry: fatal error
  if (flags_id & LE_FATAL_ERROR)
    m_fatal_error = TRUE;

  DWORD other_flags = flags_id & (255 ^ LE_FATAL_ERROR);

  if (!other_flags)
    return TRUE;

  // get selected entry, supose is the right entry - the exception one
  DWORD last_flag = 0;
  char* label = get_selected_entry ( put_in last_flag );

  // test if the selected entry has the right label
  if ( ! label || strcmp (label, LH_EXCEPTION) != 0 )
    // get exception looking for it in history
    label = get_entry ( 0, LH_EXCEPTION, FALSE, put_in last_flag );

  // setup flag
  DWORD new_flag = last_flag | other_flags;

  class_string new_lab;
  char* lab = new_lab.duplicate (LH_EXCEPTION);
  if (! lab ) {
    m_fatal_error = TRUE;
    return FALSE;
  }

  BOOL res = update ( lab, get_from new_flag, TRUE );
  if (!res) m_fatal_error = TRUE;

  return res;
}
